from ultimapy.settings import ultima_file_path
from PIL import Image
from .utils import read_byte, get_arbg_from_16_bit


class ASCIIFont:
    """
        Results 100% accurate to C# Ultima SDK in the limited testing done.
        Some methods not implemented.
        Should be reconfigured to auto load. Should never need to instantiate anything manually.
    """
    FONTS = []

    def __init__(self, buffer):
        self.header = ord(buffer.read(1))
        self.height = 0
        self.characters = {} # letter, image
        ASCIIFont.FONTS.append(self)

    @classmethod
    def load(cls):
        with open(ultima_file_path('fonts.mul'), 'rb') as f:
            for i in range(10):
                font = ASCIIFont(f)
                for k in range(224):
                    width = read_byte(f)
                    height = read_byte(f)
                    read_byte(f)  # unknown, discard
                    if k < 96:
                        font.height = max(font.height, height)
                    img = Image.new("RGBA", (width, height), (0, 0, 0, 0))
                    pos = f.tell()
                    stride = 2 * width  # 2 bytes, width
                    for y in range(height):
                        f.seek(pos)
                        pos += stride
                        for x in range(width):
                            first_byte = read_byte(f)
                            second_byte = read_byte(f)
                            pixel_num = first_byte | (second_byte << 8)
                            if pixel_num > 0:
                                img.putpixel((x, y), get_arbg_from_16_bit(pixel_num ^ 0x8000))
                    font.add_character(chr(k + 32), img)

    def add_character(self, letter, character):
        self.characters[letter] = character

    def get_character(self, letter):
        return self.characters[letter]

    def character_image_list(self, full_string):
        return [self.characters[letter] for letter in full_string]

    def get_width(self, full_string):
        img_list = self.character_image_list(full_string)
        total_width = sum(img.width for img in img_list)
        return total_width + 2

    def get_string_image(self, full_string):
        width = self.get_width(full_string)
        height = self.height + 2
        new_im = Image.new('RGBA', (width, height))
        x_offset = 2
        for im in self.character_image_list(full_string):
            new_im.paste(im, (x_offset, height - im.size[1]))
            x_offset += im.size[0]
        return new_im


ASCIIFont.load()



from . import animations
from . import art
from . import ascii_font
from . import bodies
from . import gumps
from . import hues
from . import skills

__all__ = ['animations', 'art', 'ascii_font', 'bodies', 'gumps', 'hues', 'skills']



from .file_index import FileIndex
from .utils import get_arbg_from_16_bit
from struct import unpack

from PIL import Image


class Art:
    _file_index = FileIndex("artidx.mul", "art.mul", 0x10000, 4)
    idx_length = 0
    _max_item_id = 0
    is_uoahs = False
    _cache = [None] * 0xFFFF
    _removed = [False] * 0xFFFF
    _patched = {}
    _modified = False

    class CheckSums:
        checksum = []
        pos = 0
        length = 0
        index = 0

    checksums_land = []
    checksums_static = []

    @classmethod
    def load(cls):
        cls.idx_length = cls._file_index.index_length / 12
        cls.is_uoahs = cls.idx_length >= 0x13FDC
        cls._max_item_id = 0xFFF if cls.idx_length >= 0x13FDC else 0x7FFF if cls.idx_length == 0xC000 else 0x3FFF

    @classmethod
    def get_legal_item_id(cls, item_id, check_max_id=True):
        if item_id < 0:
            return 0
        if check_max_id:
            if item_id > cls._max_item_id:
                return 0
        return item_id

    @classmethod
    def replace_static(cls, index, bitmap):
        index = cls.get_legal_item_id(index)
        index += 0x4000
        cls._cache[index] = bitmap
        cls._removed[index] = False
        if index in cls._patched:
            del cls._patched[index]
        cls._modified = True

    @classmethod
    def replace_land(cls, index, bitmap):
        index &= 0x3FFF  # 14 bits
        cls._cache[index] = bitmap
        cls._removed[index] = False
        if index in cls._patched:
            del cls._patched[index]
        cls._modified = True

    @classmethod
    def remove_static(cls, index):
        index = cls.get_legal_item_id(index)
        index += 0x4000
        cls._removed[index] = True
        cls._modified = True

    @classmethod
    def remove_land(cls, index):
        index &= 0x3FFF
        cls._removed[index] = True
        cls._modified = True

    @classmethod
    def is_valid_static(cls, index):
        index = cls.get_legal_item_id(index)
        index += 0x4000
        if cls._removed[index]:
            return False
        if cls._cache[index]:
            return True
        stream, length, extra, patched = cls._file_index.seek(index)
        if not stream:
            return False
        stream.seek(4)
        dat = unpack('H' * 4, stream.read(2) * 4)
        # todo: really need to check the output. c# appears to check a byte past the end of the array.
        # todo: also it seems to read 1 byte shorts?
        for short in dat:
            if short <= 0:
                return False
        return True

    @classmethod
    def is_valid_land(cls, index):
        index &= 0x3FFF
        if cls._removed[index]:
            return False
        if cls._cache[index]:
            return True
        return cls._file_index.valid(index)

    @classmethod
    def get_land(cls, index):
        """
        :return: bitmap, patched (bool)
        """
        index &= 0x3FFF
        patched = index in cls._patched
        if cls._removed[index]:
            return None, patched
        cache = cls._cache[index]
        if cache:
            return cache, patched
        stream, length, extra, patched = cls._file_index.seek(index)
        if not stream:
            return None, patched
        if patched:
            cls._patched[index] = True
        caching_data = True  # todo: c# this is Files.CacheData, but it's a static bool that's always true
        if caching_data:
            cls._cache[index] = cls.load_land(stream)
            return cls._cache[index], patched
        return cls.load_land(stream), patched

    @classmethod
    def get_raw_land(cls, index):
        """
            Must be used internally by UOFiddler?
        :return: bytes --
        """
        index &= 0x3FFF
        stream, length, extra, patched = cls._file_index(index)
        if not stream:
            return None
        return stream.read(length)

    @classmethod
    def get_static(cls, index, check_max_id=True):
        """
            todo: this method is almost exactly the same as get_land
        :return: bitmap, patched (bool)
        """
        index = cls.get_legal_item_id(index, check_max_id)
        index += 0x4000
        patched = index in cls._patched
        if cls._removed[index]:
            return None, patched
        cache = cls._cache[index]
        if cache:
            return cache
        stream, length, extra, patched = cls._file_index.seek(index)
        if not stream:
            return None, patched
        if patched:
            cls._patched[index] = True
        caching_data = True  # todo: c# this is Files.CacheData, but it's a static bool that's always true
        if caching_data:
            cls._cache[index] = cls.load_static(stream)
            return cls._cache[index]
        return cls.load_static(stream)

    @classmethod
    def get_raw_static(cls, index):
        """
            Must be used internally by UOFiddler?
        :return: bytes --
        """
        index = cls.get_legal_item_id(index)
        index += 0x4000
        stream, length, extra, patched = cls._file_index.seek(index)
        if not stream:
            return None
        return stream.read(length)

    # measure(bitmap) not implemented - Must be used internally by UOFiddler

    @classmethod
    def load_static(cls, stream):
        def read_number():
            return unpack('h', stream.read(2))[0]

        orig_pos = stream.tell()
        stream.seek(4, 1)

        width, height = read_number(), read_number()
        if width == 0 or height == 0:
            return None
        lookups = [0] * height
        for i in range(height):
            lookups[i] = (height + 4 + read_number()) * 2

        bitmap = Image.new("RGBA", (width, height))

        for y in range(height):
            stream.seek(orig_pos + lookups[y])
            x_offset, x_run = read_number(), read_number()
            x = 0
            while x_offset + x_run != 0:
                if x_offset > width:
                    break
                x += x_offset
                if x_offset + x_run > width:
                    break
                for pix in range(x_run):
                    val = read_number() ^ 0x8000
                    bitmap.putpixel((x + pix, y), get_arbg_from_16_bit(val))
                x += x_run
                x_offset, x_run = read_number(), read_number()
        return bitmap

    @classmethod
    def load_land(cls, stream):
        def read_number(byte_length):
            if byte_length == 1:
                return ord(stream.read(1))
            return unpack('h', stream.read(2))[0]

        bitmap = Image.new("RGBA", (44, 44))

        x_offset = 21
        x_run = 2
        for y in range(22):
            for x in range(x_run):
                bitmap.putpixel((x + x_offset, y), get_arbg_from_16_bit(read_number(2) | 0x8000))
            x_offset -= 1
            x_run += 2

        x_offset = 0
        x_run = 44
        for y in range(22):
            for x in range(x_run):
                bitmap.putpixel((x + x_offset, y + 22), get_arbg_from_16_bit(read_number(2) | 0x8000))
            x_offset += 1
            x_run -= 2

        return bitmap


if Art.idx_length == 0:
    print("Loading art..")
    Art.load()



from ultimapy.settings import ultima_file_path
from io import BytesIO
from struct import unpack
from .verdata import Verdata


class FileIndex:
    """
        - No current UOP support.
        - Patches not tested.
        - Output unconfirmed to be accurate.
        - Seek method not implemented
        - Valid() method not implemented
        - Should be refactored -- files can be non existent.
    """
    def __init__(self, idx_filename, mul_filename, length, file_idx):
        try:
            index_file = None
            self.index_file_path = ultima_file_path(idx_filename)
            index_file = open(self.index_file_path, 'rb')
            self.mul_file_path = ultima_file_path(mul_filename)
            mul_file = open(self.mul_file_path, 'rb')
        except FileNotFoundError:
            print(f"No file for index {idx_filename if not index_file else mul_filename}")
            return
        idx_file_bytes = len(index_file.read())
        index_file.seek(0)
        count = int(idx_file_bytes / 12)
        length = length or count
        self.index_length = idx_file_bytes
        self.stream = mul_file
        self.index = []
        for i in range(count):
            self.index.append(Entry3D(*unpack('i'*3, index_file.read(4*3))))
        for i in range(count, length):
            self.index.append(Entry3D(-1, -1, -1))
        patches = Verdata.patches
        if file_idx > -1:
            for idx, patch in enumerate(patches):
                if patch.file == file_idx and 0 < patch.index < length:
                    self.index[patch.index].lookup = patch.lookup
                    self.index[patch.index].length = patch.length | (1 << 31)
                    self.index[patch.index].extra = patch.extra

    def seek(self, index, is_validation=False):
        null_return = None, 0, 0, False
        if index >= len(self.index) or index < 0:
            return null_return
        entry = self.index[index]
        if entry.lookup < 0 or entry.length < 0:
            return null_return

        length = entry.length & 0x7FFFFFFF
        extra = entry.extra
        patched = False

        if (entry.length & (1 << 31)) != 0:
            patched = True
            stream = Verdata.FILE
            stream.seek(entry.lookup)
            return BytesIO(stream.read(length)), length, extra, patched

        stream = self.stream
        if not self.stream:# or self.index_length < entry.lookup:
            return null_return

        byte_stream = None  # return nothing if is validation
        if not is_validation:
            stream.seek(entry.lookup)
            byte_stream = BytesIO(stream.read(length))
        return byte_stream, length, extra, patched

    def valid(self, index):
        stream, length, extra, patched = self.seek(index, is_validation=True)
        return stream is not None, length, extra, patched


class Entry3D:
    def __init__(self, lookup, length, extra):
        self.lookup = lookup
        self.length = length
        self.extra = extra



from ultimapy.sdk.file_index import FileIndex
from ultimapy.sdk.hues import Hues
from PIL import Image
from .utils import get_arbg_from_16_bit, trim
from struct import unpack
from ultimapy.sdk.tile_data import item_data


class Gumps:
    _file_index = FileIndex("gumpidx.mul", "gumpart.mul", 0xFFFF, 12)
    _cache = [None] * 0xFFFF
    _removed = [False] * 0xFFFF
    _patched = {}

    @classmethod
    def get_gump(cls, index, hue=0, partial_hue=False):
        def read_number():
            return unpack('h', stream.read(2))[0]
        # todo: patching etc
        patched = False
        stream, length, extra, patched = cls._file_index.seek(index)
        if not stream or extra == -1:
            return
        width = (extra >> 16) & 0xFFFF
        height = extra & 0xFFFF
        if not width or not height:
            return

        orig_pos = stream.tell()
        lookups = [0] * height
        for i in range(height):
            lookups[i] = unpack('i', stream.read(4))[0] * 4

        bitmap = Image.new("RGBA", (width, height))
        for y in range(height):
            x = 0
            stream.seek(orig_pos + lookups[y])
            while x < width:
                color, x_run = read_number(), read_number()
                if color != 0:
                    color ^= 0x8000
                    for pix in range(x_run):
                        bitmap.putpixel((x + pix, y), get_arbg_from_16_bit(color))
                x += x_run
        if hue:
            hue = (hue & 0x3FFF) - 1
            return Hues.HUES[hue].apply_to(bitmap, only_grey_pixels=partial_hue)
        return bitmap

    def paperdoll_of(self, female, body_hue, layers, order=True):
        """
            Draw a paperdoll of the passed layers.
            :param female: Avatar is female?
            :param body_hue: Body hue of the player.
            :param layers: List of tuples in the format (ItemID, Hue).
            :param order: Whether to try order them to look correct. If set to false, the list of tuples must be in the
                          order in which you want them rendered. This would be useful if the default ordering does not contain
                          the items in your list.
            :return: A trimmed image of all the layers.
            """
        img = Image.new("RGBA", (260, 237))
        body = 12 + int(female)
        body_img = get_gump(body, body_hue)
        img.paste(body_img)

        for item_id, hue in layers:
            item = item_data(item_id)
            anim = item.animation
            partial = item.partial_hue
            item = get_gump(anim + 50000 + int(female) * 10000, hue, partial) or get_gump(anim + 50000, hue, partial)
            if not item:
                continue
            img.paste(item, (0, 0), item)
        trimmed = trim(img)
        return trimmed


get_gump = Gumps.get_gump



from ultimapy.settings import ultima_file_path
from .file_index import FileIndex
from struct import unpack, error


class SkillGroup:
    categories = []
    skill_list = []
    unicode = False

    @classmethod
    def load(cls):
        def read_char():
            return unpack('c', f.read(1))[0].decode('cp1252') if not cls.unicode else chr(unpack('h', f.read(2))[0])
        Skills.load()
        with open(ultima_file_path('skillgrp.mul'), 'rb') as f:
            start = 4
            strlen = 17
            count = unpack('i', f.read(4))[0]
            if count == -1:
                cls.unicode = True
                count = unpack('i', f.read(4))[0]
                start *= 2
                strlen *= 2

            cls.categories.append("Misc")
            for i in range(count - 1):
                f.seek(start + (i * strlen))
                name = ''
                c = read_char()
                while c != '\x00':
                    name += c
                    c = read_char()
                cls.categories.append(name)
            f.seek(start + (count-1)*strlen)
            skill_num = unpack('i', f.read(4))[0]
            for i, skill in enumerate(Skills.entries):
                skill.category = cls.categories[skill_num]
                try:
                    skill_num = unpack('i', f.read(4))[0]
                except error:
                    break


class Skills:
    entries = []
    @classmethod
    def load(cls):
        cls.file_index = FileIndex(ultima_file_path('Skills.idx'), ultima_file_path('skills.mul'), None, 16)
        for i in range(cls.file_index.index_length):
            stream, length, extra, patched = cls.file_index.seek(i)
            if stream is None:
                continue
            is_action = unpack('?', stream.read(1))[0]
            name = unpack(f'{(length-1)}s', stream.read(length-1))[0].decode('cp1252')[:-1]
            cls.entries.append(SkillInfo(i, name, is_action, extra))


class SkillInfo:
    def __init__(self, num, name, is_action, extra):
        self.index = num
        self.name = name or ''
        self.is_action = is_action
        self.extra = extra
        self.category = None


if not getattr(Skills, 'file_index', False):
    print("Loading skills")
    SkillGroup.load()



from PIL import Image, ImageChops
from ultimapy.settings import ULTIMA_MOUNT_IDS


def paste_centered(parent_image, input_frame, frame_no=0):
    """
    :param parent_image: bitmap to be pasted onto
    :param input_frame: ultimapy.animations.Frame
    :param frame_no: the frame number
    """
    if not input_frame or len(input_frame) <= frame_no:
        return
    input_frame = input_frame[frame_no]
    new_center = (100 - input_frame.center[0], 100 - input_frame.center[1] - input_frame.bitmap.size[1])
    parent_image.paste(input_frame.bitmap, new_center, input_frame.bitmap)


def trim(im):
    """Trim an image of surrounding transparency"""
    bg = Image.new(im.mode, im.size, im.getpixel((0, 0)))
    diff = ImageChops.difference(im, bg)
    diff = ImageChops.add(diff, diff, 2.0, -100)
    bbox = diff.getbbox()
    if bbox:
        return im.crop(bbox)


def read_byte(buf):
    """Should be removed."""
    return ord(buf.read(1))


def safe_list_get(lst, idx, default):
    try:
        return lst[idx]
    except IndexError:
        return default


def get_arbg_from_16_bit(sixteen):
    """Should be refactored."""
    def get_arbg(num):
        num = num * 255 / 31
        num = round(num)
        return num
    blue = get_arbg(sixteen & 31)
    green = get_arbg((sixteen >> 5) & 31)
    red = get_arbg((sixteen >> 10) & 31)
    alpha = ((sixteen >> 15) & 1) * 255
    return red, green, blue, alpha


def is_mount(item_id):
    return int(item_id) in ULTIMA_MOUNT_IDS


def mount_info(layers):
    for item_id, hue in layers:
        if is_mount(item_id):
            return item_id, hue
    return None, None


from ultimapy.settings import ultima_file_path

from struct import unpack


class Verdata:
    """
        No reason to instantiate.
        Assumed to be working as per C# Ultima SDK but untested.
    """
    FILE = open(ultima_file_path('verdata.mul'), 'rb')
    patches = []

    @classmethod
    def load(cls):
        total_patches = unpack('i', cls.FILE.read(4))[0]
        for i in range(total_patches):
            cls.patches.append(Entry5D(*unpack('i'*5, cls.FILE.read(20))))

    @classmethod
    def seek(cls, lookup):
        cls.FILE.Seek(lookup)


class Entry5D:
    def __init__(self, file, index, lookup, length, extra):
        self.file = file
        self.index = index
        self.lookup = lookup
        self.length = length
        self.extra = extra


if len(Verdata.patches) == 0:
    print("Loading Verdata")
    Verdata.load()



import os

try:
    from django.conf import settings
except ModuleNotFoundError:
    settings = None

# if an environment file exists, load it into the env.
# if django exists, load the variables from the django config over the environment variable.
if os.path.exists('environment.ini'):
    env_file = open('environment.ini')
    for line in env_file:
        trim_line = line.rstrip('\n').strip()
        if not trim_line or trim_line.startswith("#"):
            continue
        if "=" not in trim_line:
            print(f"Skipping invalid environment.ini variable: '{trim_line}'")
        variable, value = trim_line.split('=')
        os.environ[variable] = value


def get_django_var_or_env(variable_name):
    return getattr(
        settings,
        variable_name,
        os.environ.get(variable_name, None)
    )


def ultima_file_path(file_name):
    return os.path.join(ULTIMA_FILES_DIR, file_name)


ULTIMA_MOUNT_IDS = get_django_var_or_env('MOUNT_IDS')
if type(ULTIMA_MOUNT_IDS) is str:
    import json
    ULTIMA_MOUNT_IDS = json.loads(ULTIMA_MOUNT_IDS)

if not ULTIMA_MOUNT_IDS:
    ULTIMA_MOUNT_IDS = [
        228, 200, 218, 204, 179, 226, 219, 116, 178, 220, 210, 117
    ]

ULTIMA_FILES_DIR = os.path.abspath(get_django_var_or_env('ULTIMA_FILES_DIR'))



from ultimapy.settings import ultima_file_path

import functools
from struct import unpack
from .art import Art


class TileData:
    loaded = False
    land_header = [0] * 512
    land_data = [None] * 0x4000
    item_header = []
    item_data = []
    height_table = []
    new_format = False

    @classmethod
    def load(cls):
        cls.loaded = True
        with open(ultima_file_path('tiledata.mul'), 'rb') as f:
            f.seek(0, 2)
            length = f.tell()
            f.seek(0)
            cls.new_format = Art.is_uoahs
            j = 0
            for i in range(0, len(cls.land_data), 32):
                cls.land_header[j] = unpack('i', f.read(4))[0]
                j += 1
                for count in range(32):
                    mul_type = NewLandMul if cls.new_format else OldLandMul
                    cls.land_data[i + count] = mul_type.from_stream(f)

            remaining = length - f.tell()
            mul_type = NewItemMul if cls.new_format else OldItemMul
            cls.item_header = [0] * int(remaining / (mul_type.size * 32 + 4))
            item_length = len(cls.item_header) * 32
            cls.item_data = [None] * item_length
            cls.height_table = [0] * item_length
            j = 0
            for i in range(0, item_length, 32):
                cls.item_header[j] = unpack('i', f.read(4))[0]
                j += 1
                for count in range(32):
                    cls.item_data[i + count] = mul_type.from_stream(f)


class OldLandMul:
    def __init__(self, flags, tex_id, name):
        self.flags = flags
        self.texID = tex_id
        self.name = name.decode('cp1252')

    @classmethod
    def from_stream(cls, stream):
        return OldLandMul(*unpack('ih20s', stream.read(4 + 2 + 20)))


class NewLandMul(OldLandMul):
    def __init__(self, flags, unk1, tex_id, name):
        super().__init__(flags, tex_id, name)
        self.unk1 = unk1

    @classmethod
    def from_stream(cls, stream):
        return NewLandMul(*unpack('iih20s', stream.read(4 + 4 + 2 + 20)))


class OldItemMul:
    size = 4 + 1 * 9 + 2 * 2 + 20

    @functools.lru_cache()
    def partial_hue(self):
        return self.flags & 0x00040000

    def __init__(self, flags, weight, quality, misc, unk2, amt, anim, unk3, hue, stacking_offset, value, height, name):
        self.flags = flags
        self.weight = int(weight)
        self.quality = int(quality)
        self.misc = misc
        self.unk2 = int(unk2)
        self.quantity = int(amt)
        self.animation = anim
        self.unk3 = int(unk3)
        self.hue = int(hue)
        self.stacking_offset = int(stacking_offset)
        self.value = int(value)
        self.height = int(height)
        self.name = name.decode('cp1252')
        self.name = self.name.replace("\x00", "")

    @classmethod
    def from_stream(cls, stream):
        return OldItemMul(*unpack('ibbhbbhbbbbb20s', stream.read(cls.size)))


class NewItemMul(OldItemMul):
    size = 4 * 2 + 1 * 9 + 2 * 2 + 20

    def __init__(self, flags, unk1, weight, quality, misc, unk2, amt, anim, unk3, hue, stacking_offset, value, height, name):
        super().__init__(flags, weight, quality, misc, unk2, amt, anim, unk3, hue, stacking_offset, value, height, name)
        self.unk1 = unk1

    @classmethod
    def from_stream(cls, stream):
        return NewItemMul(*unpack('iibbhbbhbbbbb20s', stream.read(cls.size)))


if not TileData.loaded:
    TileData.load()


def item_data(index):
    return TileData.item_data[index]


def land_data(index):
    return TileData.land_data[index]



from io import BytesIO
from struct import unpack

from PIL import Image

from .bodies import *
from .file_index import FileIndex
from .tile_data import item_data
from .hues import Hues
from .utils import get_arbg_from_16_bit, mount_info, paste_centered, trim, is_mount


class Animation:
    _table = None  # integers
    _file_indexes = [
        FileIndex("anim.idx", "anim.mul", 0x40000, 6),
        FileIndex("anim2.idx", "anim2.mul", 0x10000, -1),  # i don't have these files to test anything
        FileIndex("anim3.idx", "anim3.mul", 0x20000, -1),  # i don't have these files to test anything
        FileIndex("anim4.idx", "anim4.mul", 0x20000, -1),
        FileIndex("anim5.idx", "anim5.mul", 0x20000, -1)
    ]

    @classmethod
    def get_animation(cls, body, action, direction, hue, first_frame, partial_hue=False):
        body, hue = cls.translate(body, hue)
        body, file_type = BodyConverter.convert(body)
        file_index, index = cls.get_file_index(body, action, direction, file_type)
        stream, length, extra, patched = file_index.seek(index)
        if not stream:
            return None
        flip = direction > 4
        palette = [x ^ 0x8000 for x in unpack('H' * 256, stream.read(2 * 256))]
        start = stream.tell()
        frame_count = unpack('i', stream.read(4))[0]
        frames = [None] * frame_count
        lookups = [start + x for x in unpack('i' * frame_count, stream.read(4 * frame_count))]
        hue = (hue & 0x3FFF) - 1
        hue_obj = None
        if 0 <= hue < len(Hues.HUES):
            hue_obj = Hues.HUES[hue]
        if first_frame:
            frame_count = 1
        for i in range(frame_count):
            stream.seek(lookups[i])
            frames[i] = Frame(palette, stream, flip)
            if hue_obj and frames[i] and frames[i].bitmap:
                hue_obj.apply_to(frames[i].bitmap, partial_hue)
        return frames

    @classmethod
    def load_table(cls):
        count = int(400 + (cls._file_indexes[0].index_length - 35000) / 175)
        table = cls._table = [None] * count
        for i in range(count):
            bte = BodyTable.entries.get(i, None)
            if bte is None or BodyConverter.contains(i):
                table[i] = i
            else:
                table[i] = bte.old_id | (1 << 31) | ((bte.new_hue & 0xFFFF) << 15)

    @classmethod
    def translate(cls, body, hue=None):
        if cls._table is None:
            cls.load_table()
        if body <= 0 or body >= len(cls._table):
            return 0, 0
        t = cls._table[body]  # except?
        if hue is None:
            return t & 0x7FFF, 0
        if t & (1 << 31) != 0:
            body = t & 0x7FFF
            vhue = (hue & 0x3FFF) - 1
            if vhue < 0 or vhue >= len(Hues.HUES):
                hue = (t >> 15) & 0xFFFF
        return body, hue

    @classmethod
    def is_action_defined(cls, body, action, direction):
        body = cls.translate(body)
        body, file_type = BodyConverter.convert(body)
        file_index, index = cls.get_file_index(body, action, direction, file_type)
        valid, length, _, _ = file_index.valid(index)
        return valid and length > 0

    @classmethod
    def is_anim_defined(cls, body, action, direction, file_type):
        file_index, index = cls.get_file_index(body, action, direction, file_type)
        stream, length, _, _ = file_index.seek(index)  # todo: why seek? is it used?
        return stream and length > 0

    @classmethod
    def get_anim_count(cls, file_type):
        if file_type == 2:
            info = [200, cls._file_indexes[file_type-1].index_length - 22000 * 12, 12 * 65]
        else:
            if file_type not in range(6):
                file_type = 1
            info = [400, cls._file_indexes[file_type-1].index_length - 35000 * 12, 12 * 175]
        return info[0] + info[1] / info[2]

    @classmethod
    def get_anim_length(cls, body, file_type):
        if file_type not in [1, 2, 3, 4, 5]:
            file_type = 1
        if file_type in [1,4,5]:
            return 22 if body < 200 else 13 if body < 400 else 35
        if file_type == 2:
            return 22 if body < 200 else 13
        if file_type == 3:
            return 13 if body < 300 else 22 if body < 400 else 35
        return 0

    @classmethod
    def get_file_index(cls, body, action, direction, file_type):
        if file_type not in [1,2,3,4,5]:
            file_type = 1
        file_idx = cls._file_indexes[file_type-1]
        index = None
        if file_type == 1:
            if body < 200:
                index = body * 110
            elif body < 400:
                index = 22000 + (body - 200) * 65
            else:
                index = 35000 + (body - 400) * 175
        elif file_type == 2:
            if body < 200:
                index = body * 110
            else:
                index = 22000 + (body - 200) * 65
        elif file_type == 3:
            if body < 300:
                index = body * 65
            elif body < 400:
                index = 33000 + (body - 300) * 110
            else:
                index = 35000 + (body - 400) * 175
        elif file_type == 4:
            if body < 200:
                index = body * 110
            elif body < 400:
                index = 22000 + (body - 200) * 65
            else:
                index = 35000 + (body - 400) * 175
        elif file_type == 5:
            if body < 200 and body != 34:
                index = body * 110
            elif body < 400:
                index = 22000 + (body - 200) * 65
            else:
                index = 35000 + (body - 400) * 175

        index += action * 5

        if direction <= 4:
            index += direction
        else:
            index += direction - (direction - 4) * 2

        return file_idx, index

    @classmethod
    def get_file_name(cls, body):
        body = cls.translate(body)
        file_type = BodyConverter.convert(body)
        if file_type == 1:
            return "anim.mul"
        return f"anim{file_type}.mul"

    @classmethod
    def draw_player(cls, female, body_hue, layers, order=True):
        """
        Draw a frame of the passed layers.
        :param female: Avatar is female?
        :param body_hue: Body hue of the player.
        :param layers: List of tuples in the format (ItemID, Hue).
        :param order: Whether to try order them to look correct. If set to false, the list of tuples must be in the
                      order in which you want them rendered. This would be useful if the default ordering does not contain
                      the items in your list.
        :return: A trimmed image of all the layers.
        """
        if not layers:
            return
        mount, mount_hue = mount_info(layers)
        action = 4 if not mount else 25
        img = Image.new("RGBA", (200, 200))
        if mount:
            mount_img = Animation.get_animation(mount, 2, 1, mount_hue, True)
            paste_centered(img, mount_img)
        player_frames = Animation.get_animation(401 if female else 400, action, 1, body_hue, True)

        paste_centered(img, player_frames)
        for item_id, hue in filter(lambda layer: not is_mount(layer[0]), layers):
            item = item_data(item_id)
            clothing_frame = Animation.get_animation(item.animation, action, 1, hue, True, item.partial_hue)
            if not clothing_frame:
                continue
            paste_centered(img, clothing_frame)
        return trim(img)


def integer(binary_string):
    return int(binary_string, 2)


class Frame:
    _double_xor = (0x200 << 22) | (0x200 << 12)
    empty = None

    def __init__(self, palette, reader, flip):
        if reader is None:
            self.center = None
            self.bitmap = Image.new('RGBA', (0, 0))
            return
        x_center, y_center, width, height = unpack('h' * 4, reader.read(2 * 4))
        if width == 0 or height == 0:
            return

        self.bitmap = Image.new('RGBA', (width, height))
        x_base = x_center - 0x200
        y_base = y_center + height - 0x200
        header = unpack('i', reader.read(4))[0]
        while header != 0x7FFF7FFF:
            header ^= self._double_xor
            bytes_to_write = header & 0xFFF
            x = ((header >> 22) & 0x3FF) + x_base
            y = ((header >> 12) & 0x3FF) + y_base
            for i in range(bytes_to_write):
                c = palette[ord(reader.read(1))]
                if c != 0:
                    self.bitmap.putpixel((x + i, y), get_arbg_from_16_bit(c))

            header = unpack('i', reader.read(4))[0]
        if flip:
            self.bitmap = self.bitmap.transpose(Image.FLIP_LEFT_RIGHT)
            x_center = width - x_center

        self.center = (x_center, y_center)


# initialize
if not Frame.empty:
    Frame.empty = Frame(None, None, None)



from .utils import safe_list_get
from ultimapy.settings import ultima_file_path


class BodyConverter:
    """
        Pretty much copied from C# Ultima SDK but could do with some refactoring.
        No known issues.
        Untested, should compare output directly with Ultima SDK.
    """
    loaded = False
    TABLE_1 = []
    TABLE_2 = []
    TABLE_3 = []
    TABLE_4 = []

    @classmethod
    def load(cls):
        print("Loading BodyConverter")
        f = open(ultima_file_path('Bodyconv.def'), 'r')
        max1 = max2 = max3 = max4 = 0
        list1 = []
        list2 = []
        list3 = []
        list4 = []
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            fields = line.split("\t")
            try:
                orig = int(safe_list_get(fields, 0, -1))
                anim2 = int(safe_list_get(fields, 1, -1))
                anim3 = int(safe_list_get(fields, 2, -1))
                anim4 = int(safe_list_get(fields, 3, -1))
                anim5 = int(safe_list_get(fields, 4, -1))
            except:
                continue
            if anim2 >= 0:
                anim2 = 122 if anim2 == 68 else anim2
                max1 = max(max1, orig)
                list1.append(orig)
                list1.append(anim2)
            if anim3 >= 0:
                max2 = max(max2, orig)
                list2.append(orig)
                list2.append(anim3)
            if anim4 >= 0:
                max3 = max(max3, orig)
                list3.append(orig)
                list3.append(anim4)
            if anim5 >= 0:
                max4 = max(max4, orig)
                list4.append(orig)
                list4.append(anim5)
        cls.TABLE_1 = [-1 for i in range(max1 + 1)]
        cls.TABLE_2 = [-1 for i in range(max2 + 1)]
        cls.TABLE_3 = [-1 for i in range(max3 + 1)]
        cls.TABLE_4 = [-1 for i in range(max4 + 1)]
        for i in range(0, len(list1), 2):
            cls.TABLE_1[list1[i]] = list1[i + 1]
        for i in range(0, len(list2), 2):
            cls.TABLE_2[list2[i]] = list2[i + 1]
        for i in range(0, len(list3), 2):
            cls.TABLE_3[list3[i]] = list3[i + 1]
        for i in range(0, len(list4), 2):
            cls.TABLE_4[list4[i]] = list4[i + 1]
        cls.loaded = True

    @classmethod
    def contains(cls, body):
        if body < 0:
            return False
        for table in [cls.TABLE_1, cls.TABLE_2, cls.TABLE_3, cls.TABLE_4]:
            if safe_list_get(table, body, -1) >= 0:
                return True
        return False

    @classmethod
    def convert(cls, body):
        if body < 0:
            return False
        for idx, table in enumerate([cls.TABLE_1, cls.TABLE_2, cls.TABLE_3, cls.TABLE_4]):
            new_body = safe_list_get(table, body, -1)
            if new_body != -1:
                return new_body, idx + 1
        return body, 1  # todo: default to returning orig body + 1 file idx, is it correct?

    @classmethod
    def get_true_body(cls, file_type, index):
        if file_type > 5 or file_type < 0 or index < 0:
            return -1
        if file_type == 1:
            return index
        try:
            return [cls.TABLE_1, cls.TABLE_2, cls.TABLE_3, cls.TABLE_4][file_type - 2].index(index)
        except ValueError:
            pass
        return -1


class BodyTable:
    entries = {}

    @classmethod
    def load(cls):
        print("Loading BodyTable")
        f = open(ultima_file_path('Body.def'), 'r')
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            # ultima sdk try catches this, but there are no exceptions thrown
            index1 = line.index("{")
            index2 = line.index("}")
            param1 = line[:index1].strip()
            param2 = line[index1+1:index2]
            param3 = line[index2+1:].strip()
            if "," in param2:
                param2 = param2[:param2.index(",")].strip()
            cls.entries[int(param1)] = BodyTableEntry(int(param2), int(param1), int(param3))


class BodyTableEntry:
    def __init__(self, old_id, new_id, new_hue):
        self.old_id = old_id
        self.new_id = new_id
        self.new_hue = new_hue


if not BodyTable.entries:
    BodyTable.load()
if not BodyConverter.loaded:
    BodyConverter.load()


from ultimapy.settings import ultima_file_path

from .utils import get_arbg_from_16_bit
from struct import unpack


class Hues:
    """
        Results 100% accurate to C# Ultima SDK in the limited testing done.
        Not all methods implemented.
        Should be reconfigured to auto load. Should never need to instantiate anything manually.
    """
    HUES = {}

    @classmethod
    def apply_to(cls, img, color, only_grey_pixels):
        """
            basically a static method for hueing an image.
        :param img: input image
        :param color: arbg color
        :param only_grey_pixels: looks retarded when true
        :return: hued image
        """
        width = img.size[0]
        height = img.size[1]
        stride = width * 2  # assuming always 2 bytes per pixel -- not sure yet
        for y in range(height):
            for x in range(width):
                col = img.getpixel((x,y))
                if col == (0, 0, 0, 0):
                    continue
                if col[0] == col[1] == col[2] or not only_grey_pixels:
                    img.putpixel((x,y), color)

        return img

    @classmethod
    def load(cls):
        print("Loading Hues")
        cls.HUES = {i: Hue(i) for i in range(3000)}
        f = open(ultima_file_path('hues.mul'), 'rb')
        block_count = max(375, int(unpack('i', f.read(4))[0] / 708))
        f.seek(0)
        headers = []  # why?
        idx = 0
        for i in range(block_count):
            headers.append(unpack('i', f.read(4)))
            for j in range(8):
                new_hue = Hue(idx)
                new_hue.colors = [x | 0x8000 for x in unpack('H'*32, f.read(64))]
                new_hue.table_start, new_hue.table_end = unpack('hh', f.read(4))
                new_hue.name = f.read(20).decode('cp1252')
                cls.HUES[idx] = new_hue
                idx += 1


class Hue:
    """
        Results 100% accurate to C# Ultima SDK in the limited testing done.
        Not all methods implemented.
    """
    def __init__(self, index):
        self.name = None
        self.index = index
        self.colors = [] # numbers
        self.table_start = 0
        self.table_end = 0

    def apply_to(self, img, only_grey_pixels=True):
        width = img.size[0]
        height = img.size[1]
        for y in range(height):
            for x in range(width):
                col = img.getpixel((x,y))
                if col == (0, 0, 0, 0):
                    continue
                if col[0] == col[1] == col[2] or not only_grey_pixels:
                    r = int(col[0] / 255 * 31)
                    img.putpixel((x,y), get_arbg_from_16_bit(self.colors[r]))

        return img

if not Hues.HUES:
    Hues.load()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ultimapy",
    version="0.0.11",
    author="Jack Ward",
    author_email="jackward84@gmail.com",
    description="UltimaPy - Extract information and images from the UO client files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jackuoll/ultima-py",
    packages=setuptools.find_packages(),
    license="Beerware",
    install_requires=[
        'Pillow==5.2.0',
        'imageio==2.3.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
